"""Module responsible for building outbound Cync TCP packets."""

import threading

from . import inner_packet_builder
from .packet import MessageType, PipeDirection, generate_zero_bytes

PROTOCOL_VERSION = 3

_packet_counter = 1
_packet_counter_lock = threading.Lock()


def build_login_request_packet(authorize_string: str, user_id: int):
    version = PROTOCOL_VERSION

    version_byte = version.to_bytes(1, "big")
    user_id_bytes = user_id.to_bytes(4, "big")
    user_auth_length_bytes = len(authorize_string).to_bytes(2, "big")
    user_auth_bytes = bytes(authorize_string, "ascii")
    suffix_bytes = bytes.fromhex("00001e")

    payload = version_byte + user_id_bytes + user_auth_length_bytes + user_auth_bytes + suffix_bytes
    header = _generate_header(MessageType.LOGIN.value, False, payload)

    return header + payload


def build_state_query_request_packet(device_id: int):
    device_id_bytes = device_id.to_bytes(4, "big")
    packet_counter = _get_and_increment_packet_counter()
    packet_counter_bytes = packet_counter.to_bytes(2, "big")

    pipe_packet = inner_packet_builder.build_query_device_inner_packet(PipeDirection.REQUEST.value)

    payload = device_id_bytes + packet_counter_bytes + generate_zero_bytes(1) + pipe_packet
    header = _generate_header(MessageType.PIPE.value, False, payload)

    return header + payload


def build_power_state_request_packet(device_id: int, standalone_mesh_id: int, is_on: bool):
    device_id_bytes = device_id.to_bytes(4, "big")
    packet_counter = _get_and_increment_packet_counter()
    packet_counter_bytes = packet_counter.to_bytes(2, "big")

    pipe_packet = inner_packet_builder.build_power_state_inner_packet(PipeDirection.REQUEST.value, standalone_mesh_id,
                                                                      is_on)

    payload = device_id_bytes + packet_counter_bytes + generate_zero_bytes(1) + pipe_packet
    header = _generate_header(MessageType.PIPE.value, False, payload)

    return header + payload


def build_brightness_request_packet(device_id: int, standalone_mesh_id: int, brightness: int):
    device_id_bytes = device_id.to_bytes(4, "big")
    packet_counter = _get_and_increment_packet_counter()
    packet_counter_bytes = packet_counter.to_bytes(2, "big")

    pipe_packet = inner_packet_builder.build_brightness_inner_packet(PipeDirection.REQUEST.value, standalone_mesh_id,
                                                                     brightness)

    payload = device_id_bytes + packet_counter_bytes + generate_zero_bytes(1) + pipe_packet
    header = _generate_header(MessageType.PIPE.value, False, payload)

    return header + payload


def build_color_temp_request_packet(device_id: int, standalone_mesh_id: int, color_temp: int):
    device_id_bytes = device_id.to_bytes(4, "big")
    packet_counter = _get_and_increment_packet_counter()
    packet_counter_bytes = packet_counter.to_bytes(2, "big")

    pipe_packet = inner_packet_builder.build_color_temp_inner_packet(PipeDirection.REQUEST.value, standalone_mesh_id,
                                                                     color_temp)

    payload = device_id_bytes + packet_counter_bytes + generate_zero_bytes(1) + pipe_packet
    header = _generate_header(MessageType.PIPE.value, False, payload)

    return header + payload


def build_rgb_request_packet(device_id: int, standalone_mesh_id: int, rgb: tuple[int, int, int]):
    device_id_bytes = device_id.to_bytes(4, "big")
    packet_counter = _get_and_increment_packet_counter()
    packet_counter_bytes = packet_counter.to_bytes(2, "big")

    pipe_packet = inner_packet_builder.build_rgb_inner_packet(PipeDirection.REQUEST.value, standalone_mesh_id, rgb)

    payload = device_id_bytes + packet_counter_bytes + generate_zero_bytes(1) + pipe_packet
    header = _generate_header(MessageType.PIPE.value, False, payload)

    return header + payload


def build_probe_request_packet(device_id: int):
    device_id_bytes = device_id.to_bytes(4, "big")
    packet_counter = _get_and_increment_packet_counter()
    packet_counter_bytes = packet_counter.to_bytes(2, "big")

    payload = device_id_bytes + packet_counter_bytes + generate_zero_bytes(1) + bytearray.fromhex('02')
    header = _generate_header(MessageType.PROBE.value, False, payload)

    return header + payload


def _generate_header(message_type: int, is_response: bool, payload_bytes: bytes):
    info_byte = (message_type << 4) + PROTOCOL_VERSION
    if is_response:
        info_byte += 8  # Set bit 4 to 1 if this is a response. Bit 4 is an "is_response" flag bit.

    info_byte = info_byte.to_bytes(1, "big")
    payload_size = len(payload_bytes).to_bytes(4, "big")
    return info_byte + payload_size


def _get_and_increment_packet_counter():
    with _packet_counter_lock:
        global _packet_counter
        counter_value = _packet_counter
        _packet_counter = (_packet_counter + 1) % 65536

        return counter_value



"""Definitions for groupings used in the Cync API."""

from __future__ import annotations

from abc import ABC, abstractmethod

from .controllable import CyncControllable
from pycync.devices import CyncDevice
from pycync.exceptions import UnsupportedCapabilityError
from pycync.tcp.command_client import CommandClient
from pycync.devices.capabilities import CyncCapability


class GroupedCyncDevices(ABC):
    """Abstract definition for a Cync device grouping."""

    @abstractmethod
    def get_device_types(self) -> frozenset[type[CyncDevice]]:
        """Returns all distinct device types found in the group."""
        pass


class CyncHome:
    """Represents a "home" in the Cync app."""

    def __init__(self, name: str, home_id: int, rooms: list[CyncRoom], global_devices: list[CyncDevice]):
        self.name = name
        self.home_id = home_id
        self.rooms = rooms
        self.global_devices = global_devices

    def contains_device_id(self, device_id: int) -> bool:
        """
        Determines whether a given device ID exists in this home.
        The home is searched recursively, so each room and group within the home will be searched.
        """

        search_result = next((device for device in self.get_flattened_device_list() if device.device_id == device_id),
                             None)

        return search_result is not None

    def get_flattened_device_list(self) -> list[CyncDevice]:
        """
        Returns a flattened list of all devices in the home, across all rooms and groups.
        """

        home_devices = self.global_devices.copy()

        for room in self.rooms:
            home_devices.extend(room.devices)
            for group in room.groups:
                home_devices.extend(group.devices)

        return home_devices


class CyncRoom(GroupedCyncDevices, CyncControllable):
    """Represents a "room" in the Cync app."""

    def __init__(self, name: str, room_id: int, parent_home: CyncHome, groups: list[CyncGroup],
                 devices: list[CyncDevice], command_client: CommandClient):
        self._name = name
        self.room_id = room_id
        self.parent_home = parent_home
        self.groups = groups
        self.devices = devices
        self._command_client = command_client

    @property
    def capabilities(self) -> frozenset[CyncCapability]:
        all_capabilities = frozenset({capability for capability in CyncCapability})
        return (all_capabilities
                .intersection([frozenset(device.capabilities) for device in self.devices])
                .intersection([group.capabilities for group in self.groups]))

    @property
    def name(self) -> str:
        return self._name

    @property
    def mesh_reference_id(self) -> int:
        return self.room_id

    @property
    def unique_id(self) -> str:
        return f"{self.parent_home.home_id}-{self.room_id}"

    def supports_capability(self, capability: CyncCapability) -> bool:
        return capability in self.capabilities

    def get_device_types(self) -> frozenset[type[CyncDevice]]:
        return frozenset({type(device) for device in self.devices}).union(
            [group.get_device_types() for group in self.groups])

    async def turn_on(self):
        if not self.supports_capability(CyncCapability.ON_OFF):
            raise UnsupportedCapabilityError()

        await self._command_client.set_power_state(self, True)

    async def turn_off(self):
        if not self.supports_capability(CyncCapability.ON_OFF):
            raise UnsupportedCapabilityError()

        await self._command_client.set_power_state(self, False)

    async def set_brightness(self, brightness):
        if not self.supports_capability(CyncCapability.DIMMING):
            raise UnsupportedCapabilityError()

        await self._command_client.set_brightness(self, brightness)

    async def set_color_temp(self, color_temp):
        if not self.supports_capability(CyncCapability.CCT_COLOR):
            raise UnsupportedCapabilityError()

        await self._command_client.set_color_temp(self, color_temp)

    async def set_rgb(self, rgb: tuple[int, int, int]):
        if not self.supports_capability(CyncCapability.RGB_COLOR):
            raise UnsupportedCapabilityError()

        await self._command_client.set_rgb(self, rgb)


class CyncGroup(GroupedCyncDevices, CyncControllable):
    """Represents a "group" in the Cync app."""

    def __init__(self, name: str, group_id: int, parent_home: CyncHome, devices: list[CyncDevice],
                 command_client: CommandClient):
        self._name = name
        self.group_id = group_id
        self.parent_home = parent_home
        self.devices = devices
        self._command_client = command_client

    @property
    def capabilities(self) -> frozenset[CyncCapability]:
        all_capabilities = frozenset({capability for capability in CyncCapability})
        return all_capabilities.intersection([frozenset(device.capabilities) for device in self.devices])

    @property
    def name(self) -> str:
        return self._name

    @property
    def mesh_reference_id(self) -> int:
        return self.group_id

    @property
    def unique_id(self) -> str:
        return f"{self.parent_home.home_id}-{self.group_id}"

    def supports_capability(self, capability: CyncCapability) -> bool:
        return capability in self.capabilities

    def get_device_types(self) -> frozenset[type[CyncDevice]]:
        return frozenset({type(device) for device in self.devices})

    async def turn_on(self):
        if not self.supports_capability(CyncCapability.ON_OFF):
            raise UnsupportedCapabilityError()

        await self._command_client.set_power_state(self, True)

    async def turn_off(self):
        if not self.supports_capability(CyncCapability.ON_OFF):
            raise UnsupportedCapabilityError()

        await self._command_client.set_power_state(self, False)

    async def set_brightness(self, brightness):
        if not self.supports_capability(CyncCapability.DIMMING):
            raise UnsupportedCapabilityError()

        await self._command_client.set_brightness(self, brightness)

    async def set_color_temp(self, color_temp):
        if not self.supports_capability(CyncCapability.CCT_COLOR):
            raise UnsupportedCapabilityError()

        await self._command_client.set_color_temp(self, color_temp)

    async def set_rgb(self, rgb: tuple[int, int, int]):
        if not self.supports_capability(CyncCapability.RGB_COLOR):
            raise UnsupportedCapabilityError()

        await self._command_client.set_rgb(self, rgb)



"""Contains definitions for various Cync devices."""

from __future__ import annotations

from typing import Tuple, Any, TYPE_CHECKING

from .controllable import CyncControllable
from pycync.exceptions import UnsupportedCapabilityError
from pycync.tcp.command_client import CommandClient
from pycync.devices.capabilities import DEVICE_CAPABILITIES, CyncCapability
from pycync.devices.device_types import DEVICE_TYPES, DeviceType

if TYPE_CHECKING:
    from .groups import CyncHome


def create_device(device_info: dict[str, Any], mesh_device_info: dict[str, Any], parent_home: CyncHome,
                  command_client: CommandClient, wifi_connected: bool = False,
                  device_datapoint_data: dict[str, Any] = None) -> CyncDevice:
    device_type_id = mesh_device_info["deviceType"]

    device_type = DEVICE_TYPES.get(device_type_id, DeviceType.UNKNOWN)

    match device_type:
        case (DeviceType.LIGHT |
              DeviceType.INDOOR_LIGHT_STRIP |
              DeviceType.OUTDOOR_LIGHT_STRIP |
              DeviceType.NEON_LIGHT_STRIP |
              DeviceType.OUTDOOR_NEON_LIGHT_STRIP |
              DeviceType.CAFE_STRING_LIGHTS |
              DeviceType.DOWNLIGHT |
              DeviceType.UNDERCABINET_FIXTURES |
              DeviceType.LIGHT_TILE):
            return CyncLight(device_info, mesh_device_info, parent_home, command_client, wifi_connected,
                             device_datapoint_data)
        case _:
            return CyncDevice(device_info, mesh_device_info, parent_home, command_client, wifi_connected,
                              device_datapoint_data)


class CyncDevice(CyncControllable):
    """Definition for a generic Cync device, with the common attributes shared between device types."""

    def __init__(self, device_info: dict[str, Any], mesh_device_info: dict[str, Any], parent_home: CyncHome,
                 command_client: CommandClient, wifi_connected: bool, device_datapoint_data: dict[str, Any] = None):
        if device_datapoint_data is None:
            device_datapoint_data = {}
        self.is_online = device_info.get("is_online", False)
        self.wifi_connected = wifi_connected
        self.device_id = device_info.get("id")
        self.mesh_device_id = mesh_device_info.get("deviceID")
        self.isolated_mesh_id = self.mesh_device_id % parent_home.home_id
        self.parent_home = parent_home
        self._name = mesh_device_info.get("displayName")
        self.device_type = mesh_device_info.get("deviceType", DeviceType.UNKNOWN.value)
        self.mac = device_info.get("mac")
        self.product_id = device_info.get("product_id")
        self.authorize_code = device_info.get("authorize_code")
        self.datapoints = device_datapoint_data
        self._capabilities = DEVICE_CAPABILITIES.get(self.device_type, {})
        self._command_client = command_client

    def set_wifi_connected(self, wifi_connected: bool):
        self.wifi_connected = wifi_connected

    def set_datapoints(self, datapoints: dict[str, Any]):
        """Currently not used. Will be once datapoint-driven devices are implemented."""
        self.datapoints = datapoints

    @property
    def capabilities(self) -> frozenset[CyncCapability]:
        return self._capabilities

    @property
    def name(self) -> str:
        return self._name

    @property
    def mesh_reference_id(self) -> int:
        return self.isolated_mesh_id

    @property
    def unique_id(self) -> str:
        return f"{self.parent_home.home_id}-{self.device_id}"

    def supports_capability(self, capability: CyncCapability) -> bool:
        return capability in self.capabilities


class CyncLight(CyncDevice):
    """Class for representing Cync lights."""

    def __init__(self, device_info: dict[str, Any], mesh_device_info: dict[str, Any], parent_home: CyncHome,
                 command_client: CommandClient, wifi_connected: bool, device_datapoint_data: dict[str, Any] = None):
        super().__init__(device_info, mesh_device_info, parent_home, command_client, wifi_connected,
                         device_datapoint_data)

        self._is_on = False
        self._brightness = 0
        self._color_temp = 0
        self._rgb = 0, 0, 0

    @property
    def is_on(self) -> bool:
        if not self.supports_capability(CyncCapability.ON_OFF):
            raise UnsupportedCapabilityError()

        return self._is_on

    @property
    def brightness(self) -> int:
        if not self.supports_capability(CyncCapability.DIMMING):
            raise UnsupportedCapabilityError()

        return self._brightness

    @property
    def color_temp(self) -> int:
        """
        Return color temp between 1-100. Returns zero if bulb is not in color temp mode,
        or if the bulb does not have its datapoint set.
        """
        if not self.supports_capability(CyncCapability.CCT_COLOR):
            raise UnsupportedCapabilityError()

        return self._color_temp if 1 <= self._color_temp <= 100 else 0

    @property
    def color_mode(self) -> int:
        """
        A more generalized version of 'color temp', as this field is also used to store state that indicates
        when the light is in RGB mode and effect modes.
        """
        return self._color_temp

    @property
    def rgb(self) -> Tuple[int, int, int]:
        """
        Return RGB tuple, with each value between 1-256. Returns all zeros if bulb is not in RGB mode,
        or if the bulb does not have its datapoint set.
        """
        if not self.supports_capability(CyncCapability.RGB_COLOR):
            raise UnsupportedCapabilityError()

        return self._rgb

    def update_state(self, is_on: bool, brightness: int = None, color_temp: int = None, rgb: Tuple[int, int, int] = None, is_online: bool = None):
        self._is_on = is_on
        if brightness is not None:
            self._brightness = brightness
        if color_temp is not None:
            self._color_temp = color_temp
        if rgb is not None:
            self._rgb = rgb
        if is_online is not None:
            self.is_online = is_online

    async def turn_on(self):
        if not self.supports_capability(CyncCapability.ON_OFF):
            raise UnsupportedCapabilityError()

        await self._command_client.set_power_state(self, True)

    async def turn_off(self):
        if not self.supports_capability(CyncCapability.ON_OFF):
            raise UnsupportedCapabilityError()

        await self._command_client.set_power_state(self, False)

    async def set_brightness(self, brightness):
        if not self.supports_capability(CyncCapability.DIMMING):
            raise UnsupportedCapabilityError()

        await self._command_client.set_brightness(self, brightness)

    async def set_color_temp(self, color_temp):
        if not self.supports_capability(CyncCapability.CCT_COLOR):
            raise UnsupportedCapabilityError()

        await self._command_client.set_color_temp(self, color_temp)

    async def set_rgb(self, rgb: tuple[int, int, int]):
        if not self.supports_capability(CyncCapability.RGB_COLOR):
            raise UnsupportedCapabilityError()

        await self._command_client.set_rgb(self, rgb)



"""
This TCP manager is responsible for maintaining the open TCP connection against the Cync server, and sending/receiving
packets associated with device events.
"""

from __future__ import annotations
from typing import TYPE_CHECKING

import asyncio
import logging
import ssl
import struct
import threading
from typing import Callable

from pycync import User
from . import packet_builder, packet_parser
from .packet import MessageType

if TYPE_CHECKING:
    from pycync.devices import CyncDevice

TCP_API_HOSTNAME = "cm-sec.gelighting.com"
TCP_API_TLS_PORT = 23779

class TcpManager:
    _LOGGER = logging.getLogger(__name__)

    def __init__(self, user: User, client_callback: Callable):
        self._user = user

        self._client_closed = False
        self._login_acknowledged = False
        self._loop = None
        self._client_thread = threading.Thread(target=self._open_thread_connection, daemon=True, name=f"CyncTcpManager-{user.user_id}")
        self.transport = None
        self.protocol = None
        self.packet_queue = asyncio.Queue()
        self.client_callback = client_callback

        self._client_thread.start()

    def _open_thread_connection(self):
        self._loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self._loop)
        self._loop.run_until_complete(self._begin_client_loop())

    async def _begin_client_loop(self):
        while not self._client_closed:
            try:
                await self._establish_tcp_connection()
            except Exception:
                self._LOGGER.error("Failed to connect to Cync server. Retrying in 5 seconds...")
                await asyncio.sleep(5)
            else:
                process_packets = asyncio.create_task(self._process_packets(), name="Process Cync Packets")
                heartbeat_task = asyncio.create_task(self._send_pings(), name="Send Heartbeats")
                read_write_tasks = [process_packets, heartbeat_task]
                try:
                    done, pending = await asyncio.wait(read_write_tasks, return_when=asyncio.FIRST_EXCEPTION)
                    for task in done:
                        try:
                            task.result()
                        except ShuttingDown:
                            self._LOGGER.debug("Cync client shutting down")
                        except Exception as e:
                            self._LOGGER.error(e)
                    for task in pending:
                        task.cancel()
                except Exception as e:
                    self._LOGGER.error(e)
                finally:
                    self.transport.close()

                if not self._client_closed:
                    self._LOGGER.debug("Cync server connection closed. Reconnecting in 10 seconds...")
                    await asyncio.sleep(10)

    async def _establish_tcp_connection(self):
        context = ssl.create_default_context()
        try:
            self.transport, self.protocol = await self._loop.create_connection(lambda: CyncTcpProtocol(self.packet_queue, self._user), host=TCP_API_HOSTNAME, port=TCP_API_TLS_PORT, ssl=context)
        except Exception:
            # Normally this isn't something you'd want to do.
            # However, Cync's server has a 2+ year expired certificate and the common name doesn't match.
            # Why they haven't renewed/fixed it, and why their devices allow this, who knows...
            self._LOGGER.debug("Could not connect to Cync TCP server with strict TLS. Using relaxed TLS.")

            context.check_hostname = False
            context.verify_mode = ssl.CERT_NONE
            self.transport, self.protocol = await self._loop.create_connection(lambda: CyncTcpProtocol(self.packet_queue, self._user), host=TCP_API_HOSTNAME, port=TCP_API_TLS_PORT, ssl=context)

    async def _process_packets(self):
        """Process parsed packets as they're added to the async queue."""

        while True:
            parsed_packet = await self.packet_queue.get()

            match parsed_packet.message_type:
                case MessageType.LOGIN.value:
                    self._login_acknowledged = True
                case MessageType.DISCONNECT.value:
                    self._login_acknowledged = False
                    raise ConnectionClosedError

            self.client_callback(parsed_packet)

    async def _send_pings(self):
        """Periodically send a ping to the Cync server as a connection heartbeat."""

        while not self._client_closed:
            await asyncio.sleep(20)
            self._loop.call_soon_threadsafe(self._send_request, bytes.fromhex('d300000000'))
        raise ShuttingDown

    def probe_devices(self, devices: list[CyncDevice]):
        """Probe all account devices to see which ones are responsive over Wi-Fi."""

        for device in devices:
            probe_device_packet = packet_builder.build_probe_request_packet(device.device_id)
            self._loop.call_soon_threadsafe(self._send_request, probe_device_packet)

    def shut_down(self):
        """Shut down the Cync client connection."""

        self._client_closed = True
        self._loop.call_soon_threadsafe(self._send_request, bytes.fromhex('e30000000103'))

    def update_mesh_devices(self, hub_devices: list[CyncDevice]):
        """Get new device state."""
        for hub_device in hub_devices:
            state_request_packet = packet_builder.build_state_query_request_packet(hub_device.device_id)
            self._loop.call_soon_threadsafe(self._send_request, state_request_packet)

    def set_power_state(self, hub_device: CyncDevice, mesh_id: int, is_on: bool):
        """Set device(s) to either on or off."""
        request_packet = packet_builder.build_power_state_request_packet(hub_device.device_id, mesh_id, is_on)
        self._loop.call_soon_threadsafe(self._send_request, request_packet)

    def set_brightness(self, hub_device: CyncDevice, mesh_id: int, brightness: int):
        """Sets the brightness."""
        request_packet = packet_builder.build_brightness_request_packet(hub_device.device_id, mesh_id, brightness)
        self._loop.call_soon_threadsafe(self._send_request, request_packet)

    def set_color_temp(self, hub_device: CyncDevice, mesh_id: int, color_temp: int):
        """Sets the color temperature."""
        request_packet = packet_builder.build_color_temp_request_packet(hub_device.device_id, mesh_id, color_temp)
        self._loop.call_soon_threadsafe(self._send_request, request_packet)

    def set_rgb(self, hub_device: CyncDevice, mesh_id: int, rgb: tuple[int, int, int]):
        """Sets the RGB color."""
        request_packet = packet_builder.build_rgb_request_packet(hub_device.device_id, mesh_id, rgb)
        self._loop.call_soon_threadsafe(self._send_request, request_packet)

    def _send_request(self, request):
        async def send():
            while not self._login_acknowledged:
                await asyncio.sleep(1)
                self._LOGGER.debug("Awaiting login acknowledge before sending request.")
            self.transport.write(request)

        self._loop.create_task(send())

class CyncTcpProtocol(asyncio.Protocol):
    """Protocol class for processing the Cync TCP packets."""

    _LOGGER = logging.getLogger(__name__)

    def __init__(self, packet_queue: asyncio.Queue, user):
        self._transport = None
        self._packet_queue = packet_queue
        self._user = user

    def connection_made(self, transport):
        self._transport = transport

        self._log_in()

    def data_received(self, data):
        while len(data) > 0:
            packet_length = struct.unpack(">I", data[1:5])[0]
            packet = data[:packet_length + 5]
            try:
                parsed_packet = packet_parser.parse_packet(packet, self._user.user_id)
                self._packet_queue.put_nowait(parsed_packet)
            except NotImplementedError:
                # Simply ignore the packet for now
                pass
            except Exception as ex:
                self._LOGGER.error("Unhandled exception while parsing packet: {}".format(str(ex)))
            finally:
                data = data[packet_length + 5:]

    def _log_in(self):
        login_request_packet = packet_builder.build_login_request_packet(self._user.authorize, self._user.user_id)
        self._transport.write(login_request_packet)

class ConnectionClosedError(Exception):
    """Connection closed error"""

class ShuttingDown(Exception):
    """Cync client shutting down"""



"""
The main API interface for the Cync library.
Each instance of this class corresponds to one user, and all devices/homes associated with that user.
"""

import asyncio
from typing import Callable

from .auth import Auth
from .devices import create_device, CyncDevice, device_storage
from .exceptions import MissingAuthError
from .const import REST_API_BASE_URL
from pycync.devices.groups import CyncRoom, CyncGroup, CyncHome
from pycync.tcp.command_client import CommandClient


class Cync:
    def __init__(self, auth: Auth):
        """
        Initialize a Cync object.
        The static create function should be used to create a new Cync object.
        """
        if not auth.user:
            raise MissingAuthError("No logged in user exists on auth object.")
        self._auth = auth
        self._command_client = CommandClient(auth.user)

    @classmethod
    async def create(cls, auth: Auth):
        cync_api = Cync(auth)

        await cync_api.refresh_home_info()
        cync_api._command_client.start_connection()

        return cync_api

    def get_logged_in_user(self):
        """Get logged in user."""

        return self._auth.user

    async def refresh_credentials(self):
        """Refresh user credentials."""
        await self._auth.async_refresh_user_token()
        return self.get_logged_in_user()

    def set_update_callback(self, update_callback: Callable):
        """
        Set the callback function that will be called when a device's state changes,
        or when a poll request for device state receives a response.
        """
        device_storage.set_user_device_callback(self._auth.user.user_id, update_callback)

    def update_device_states(self):
        """Query the server for current device states, and update the devices."""
        asyncio.create_task(self._command_client.update_mesh_devices())

    def get_devices(self):
        """Get a flat list of devices associated with this user."""
        return device_storage.get_flattened_devices(self._auth.user.user_id)

    def get_homes(self):
        """Get all homes, devices, and groups for the account."""
        return device_storage.get_user_homes(self._auth.user.user_id)

    async def refresh_home_info(self):
        """Refresh all nested home information for this account, and update the device storage.."""
        device_info = await self._auth._send_user_request(
            f"{REST_API_BASE_URL}/v2/user/{self._auth.user.user_id}/subscribe/devices")
        home_entries = [device for device in device_info if device["source"] == 5]
        homes = []

        for home_json in home_entries:
            home_devices: list[CyncDevice] = []
            rooms: list[CyncRoom] = []
            groups: list[CyncGroup] = []
            home: CyncHome = CyncHome(home_json["name"], home_json["id"], [], [])

            mesh_device_info = await self._auth._send_user_request(
                f"{REST_API_BASE_URL}/v2/product/{home_json["product_id"]}/device/{home_json["id"]}/property")
            if "bulbsArray" in mesh_device_info:
                mesh_devices = mesh_device_info["bulbsArray"]
                for mesh_device in mesh_devices:
                    matching_device = next(device for device in device_info if device["id"] == mesh_device["switchID"])
                    created_device = create_device(matching_device, mesh_device, home, self._command_client)

                    home_devices.append(created_device)

            room_json = []
            group_json = []
            if "groupsArray" in mesh_device_info:
                room_json = [group for group in mesh_device_info["groupsArray"] if group["isSubgroup"] == False]
                group_json = [group for group in mesh_device_info["groupsArray"] if group["isSubgroup"] == True]

            for group in group_json:
                group_devices = [device for device in home_devices if
                                 device.isolated_mesh_id in group.get("deviceIDArray", [])]
                groups.append(
                    CyncGroup(group["displayName"], group["groupID"], home, group_devices, self._command_client))
                home_devices = [device for device in home_devices if device not in group_devices]

            for room in room_json:
                room_devices = [device for device in home_devices if device.isolated_mesh_id in room["deviceIDArray"]]
                room_groups = [group for group in groups if group.group_id in room.get("subgroupIDArray", [])]
                rooms.append(CyncRoom(room["displayName"], room["groupID"], home, room_groups, room_devices,
                                      self._command_client))
                home_devices = [device for device in home_devices if device not in room_devices]

            home.global_devices = home_devices
            home.rooms = rooms
            homes.append(home)

        device_storage.set_user_homes(self._auth.user.user_id, homes)

    def shut_down(self):
        """Shut down the command client instance and close its associated connections."""
        self._command_client.shut_down()



"""
Provides various definitions pertaining to a Cync TCP packet and its parsed version.
"""

from enum import Enum
from functools import reduce


class ParsedMessage:
    def __init__(self, message_type, is_response: bool, device_id, data, version, command_code=None):
        self.message_type = message_type
        self.command_code = command_code
        self.is_response = is_response
        self.version = version
        self.device_id = device_id
        self.data = data


class ParsedInnerFrame:
    def __init__(self, command_type, data):
        self.command_type = command_type
        self.data = data


class MessageType(Enum):
    LOGIN = 1
    HANDSHAKE = 2
    SYNC = 4
    PIPE = 7
    PIPE_SYNC = 8
    PROBE = 10
    PING = 13
    DISCONNECT = 14


class PipeCommandCode(Enum):
    SET_POWER_STATE = 0xd0
    SET_BRIGHTNESS = 0xd2
    SET_COLOR = 0xe2
    DEVICE_STATUS = 0xdb
    COMBO_CONTROL = 0xf0
    QUERY_DEVICE_STATUS_PAGES = 0x52


class PipeDirection(Enum):
    REQUEST = 0xf8
    RESPONSE = 0xf9
    ANNOUNCE = 0xfa


def generate_zero_bytes(length: int):
    return bytearray([0 for _ in range(length)])


def generate_checksum(byte_array: bytearray) -> int:
    return reduce(lambda acc, byte: (acc + byte) % 256, byte_array)



"""
Module containing device type related information.
All info in this module was pulled from the Cync app, and should not be manually modified
unless the changes come directly from the app's information.

Note that just because a device is listed here, does not necessarily mean this library supports it.
This is here to establish a complete list of types so implementing devices in the future has fewer steps.

List last updated from Android app version 6.20.0.54634-60b11b1f5
"""

from enum import Enum

class DeviceType(Enum):
    LIGHT = "Light"
    INDOOR_LIGHT_STRIP = "IndoorLightStrip"
    OUTDOOR_LIGHT_STRIP = "OutdoorLightStrip"
    NEON_LIGHT_STRIP = "NeonLightStrip"
    OUTDOOR_NEON_LIGHT_STRIP = "OutdoorNeonLightStrip"
    CAFE_STRING_LIGHTS = "CafeStringLights"
    DOWNLIGHT = "Downlight"
    UNDERCABINET_FIXTURES = "UndercabinetFixtures"
    LIGHT_TILE = "LightTile"
    PLUG = "Plug"
    SWITCH = "Switch"
    FAN_SPEED_SWITCH = "FanSpeedSwitch"
    WIRE_FREE_SWITCH = "WireFreeSwitch"
    WIRE_FREE_SENSOR = "WireFreeSensor"
    WIRE_FREE_REMOTE = "WireFreeRemote"
    SOL = "Sol"
    C_REACH = "C-Reach"
    THERMOSTAT = "Thermostat"
    CAMERA = "Camera"
    UNKNOWN = "Unknown"

DEVICE_TYPES = {
    # FullColorStripGen1Standalone
    8: DeviceType.INDOOR_LIGHT_STRIP,
    # CLifeA19Gen2StandaloneCECTier2
    9: DeviceType.LIGHT,
    # CSleepA19Gen2StandaloneCECTier2
    10: DeviceType.LIGHT,
    # CSleepBR30Gen2StandaloneCECTier2
    11: DeviceType.LIGHT,
    # CLifeA19TCOGen2Standalone
    13: DeviceType.LIGHT,
    # CSleepA19TCOGen2Standalone
    14: DeviceType.LIGHT,
    # CSleepBR30TCOGen2Standalone
    15: DeviceType.LIGHT,
    # CLifeA19Gen2GoogleBundle
    17: DeviceType.LIGHT,
    # CLifeA19Gen2MadeForGoogle
    18: DeviceType.LIGHT,
    # CSleepA19Gen2MadeForGoogle
    19: DeviceType.LIGHT,
    # CSleepBR30Gen2MadeForGoogle
    20: DeviceType.LIGHT,
    # FullColorA19Gen1MadeForGoogle
    21: DeviceType.LIGHT,
    # FullColorBR30Gen1MadeForGoogle
    22: DeviceType.LIGHT,
    # FullColorStripGen1MadeForGoogle
    23: DeviceType.INDOOR_LIGHT_STRIP,
    # CLifeA19Gen2MadeForGoogleCECTier2
    24: DeviceType.LIGHT,
    # CSleepA19Gen2MadeForGoogleCECTier2
    25: DeviceType.LIGHT,
    # CSleepBR30Gen2MadeForGoogleCECTier2
    26: DeviceType.LIGHT,
    # CLifeA19TCOGen2MadeForGoogle
    27: DeviceType.LIGHT,
    # CSleepA19TCOGen2MadeForGoogle
    28: DeviceType.LIGHT,
    # CSleepBR30TCOGen2MadeForGoogle
    29: DeviceType.LIGHT,
    # FullColorA19TCOGen2Standalone
    30: DeviceType.LIGHT,
    # FullColorA19TCOGen2MadeForGoogle
    31: DeviceType.LIGHT,
    # FullColorBR30TCOGen2Standalone
    32: DeviceType.LIGHT,
    # FullColorBR30TCOGen2MadeForGoogle
    33: DeviceType.LIGHT,
    # FullColorStripTCOGen2Standalone
    34: DeviceType.INDOOR_LIGHT_STRIP,
    # FullColorStripTCOGen2MadeForGoogle
    35: DeviceType.INDOOR_LIGHT_STRIP,
    # FourWireSwitchDimmerGen2
    36: DeviceType.SWITCH,
    # FourWireSwitchMotionSensingDimmerGen2
    37: DeviceType.SWITCH,
    # FourWireSwitchCircleGen2
    38: DeviceType.SWITCH,
    # FourWireSwitchPaddleGen2
    39: DeviceType.SWITCH,
    # FourWireSwitchToggleGen2
    40: DeviceType.SWITCH,
    # SingleChipFullColorUnderCabinetBarFixture12InchGen1
    41: DeviceType.UNDERCABINET_FIXTURES,
    # SingleChipFullColorUnderCabinetBarFixture18InchGen1
    42: DeviceType.UNDERCABINET_FIXTURES,
    # SingleChipFullColorUnderCabinetBarFixture24InchGen1
    43: DeviceType.UNDERCABINET_FIXTURES,
    # SingleChipFullColorUnderCabinetPuckFixtureGen1
    44: DeviceType.UNDERCABINET_FIXTURES,
    # SingleChipFullColorRecessedCanRetrofitFixture6InchGen1
    46: DeviceType.DOWNLIGHT,
    # SingleChipFullColorRecessedCanRetrofitFixture6InchGen1NonMatter
    47: DeviceType.DOWNLIGHT,
    # FourWireSwitchDimmerGen1
    48: DeviceType.SWITCH,
    # FourWireSwitchMotionSensingDimmerGen1
    49: DeviceType.SWITCH,
    # FourWireSwitchPaddleGen1
    51: DeviceType.SWITCH,
    # FourWireSwitchToggleGen1
    52: DeviceType.SWITCH,
    # FourWireSwitchCircleGen1
    53: DeviceType.SWITCH,
    # NoNeutralSwitchGen1Dimmer
    55: DeviceType.SWITCH,
    # NoNeutralSwitchGen1MotionSensingDimmer
    56: DeviceType.SWITCH,
    # NoNeutralSwitchGen1SwitchPaddle
    57: DeviceType.SWITCH,
    # NoNeutralSwitchGen1SwitchToggle
    58: DeviceType.SWITCH,
    # NoNeutralSwitchGen1SwitchCircle
    59: DeviceType.SWITCH,
    # FourWireSwitchPaddleTCOGen1
    61: DeviceType.SWITCH,
    # FourWireSwitchToggleTCOGen1
    62: DeviceType.SWITCH,
    # FourWireSwitchCircleTCOGen1
    63: DeviceType.SWITCH,
    # PlugGen1
    64: DeviceType.PLUG,
    # PlugTCOGen2
    65: DeviceType.PLUG,
    # PlugGen2
    66: DeviceType.PLUG,
    # PlugOutdoorGen1
    67: DeviceType.PLUG,
    # SingleChipPlug
    68: DeviceType.PLUG,
    # PlugOutdoorGen2
    69: DeviceType.PLUG,
    # SingleChipFullColorIndoorPremiumLightStrip16ftGen3
    71: DeviceType.INDOOR_LIGHT_STRIP,
    # SingleChipFullColorIndoorPremiumLightStrip32ftGen3
    72: DeviceType.INDOOR_LIGHT_STRIP,
    # SingleChipFullColorOutdoorNeonLightStrip16ft
    73: DeviceType.OUTDOOR_NEON_LIGHT_STRIP,
    # SingleChipFullColorOutdoorNeonLightStrip32ft
    74: DeviceType.OUTDOOR_NEON_LIGHT_STRIP,
    # SingleChipFullColorOutdoorCafeLights24ftGen1
    75: DeviceType.CAFE_STRING_LIGHTS,
    # SingleChipFullColorOutdoorCafeLights48ftGen1
    76: DeviceType.CAFE_STRING_LIGHTS,
    # SolGen1Standalone
    80: DeviceType.SOL,
    # FanSpeedSwitchGen1
    81: DeviceType.FAN_SPEED_SWITCH,
    # CLifeBR30Gen1Standalone
    82: DeviceType.LIGHT,
    # CSleepBR30Gen2Standalone
    83: DeviceType.LIGHT,
    # WireFreeMotionSensor
    96: DeviceType.WIRE_FREE_SENSOR,
    # SingleChipFullColorBulbST19Gen2
    97: DeviceType.LIGHT,
    # SingleChipFullColorBulbG25Gen2
    98: DeviceType.LIGHT,
    # SingleChipSoftWhiteBulbST19Gen2
    99: DeviceType.LIGHT,
    # SingleChipSoftWhiteBulbG25Gen2
    100: DeviceType.LIGHT,
    # SingleChipFullColorBulbBCGen1
    101: DeviceType.LIGHT,
    # SingleChipFullColorBulbBMGen1
    102: DeviceType.LIGHT,
    # SingleChipRevealSoftWhiteBulbA19Gen2
    103: DeviceType.LIGHT,
    # SingleChipFullColorBulbPAR38OutdoorGen2
    104: DeviceType.LIGHT,
    # SingleChipFullColorBulbA21Gen2
    105: DeviceType.LIGHT,
    # SingleChipRevealFullColorBulbA19Gen2
    107: DeviceType.LIGHT,
    # SingleChipRevealFullColorBulbBR30Gen2
    108: DeviceType.LIGHT,
    # SingleChipRevealFullColorBulbA21Gen2
    109: DeviceType.LIGHT,
    # SingleChipFullColorIndoorValueLightStrip16ftGen3
    110: DeviceType.INDOOR_LIGHT_STRIP,
    # PlugOutdoorGen3
    111: DeviceType.PLUG,
    # WireFreeSmartSwitchDimmer
    112: DeviceType.WIRE_FREE_SWITCH,
    # WireFreeSmartSwitchDimmerPlusColorController
    113: DeviceType.WIRE_FREE_SWITCH,
    # WireFreeSmartRemoteDimmer
    114: DeviceType.WIRE_FREE_REMOTE,
    # WireFreeSmartRemoteDimmerPlusColorController
    115: DeviceType.WIRE_FREE_REMOTE,
    # FourWireSwitchDimmerGen3
    116: DeviceType.SWITCH,
    # FourWireSwitchMotionSensingDimmerGen3
    117: DeviceType.SWITCH,
    # FourWireSwitchCircleGen3
    118: DeviceType.SWITCH,
    # FourWireSwitchPaddleGen3
    119: DeviceType.SWITCH,
    # FourWireSwitchToggleGen3
    120: DeviceType.SWITCH,
    # FanSpeedSwitchGen2
    121: DeviceType.FAN_SPEED_SWITCH,
    # ThermostatGen2
    122: DeviceType.THERMOSTAT,
    # SingleChipFullColorIndoorValueLightStrip32ftGen3
    123: DeviceType.INDOOR_LIGHT_STRIP,
    # FourWireSwitchKeypadDimmerGen1
    124: DeviceType.SWITCH,
    # FourWireSwitchPaddleDimmerGen1
    125: DeviceType.SWITCH,
    # DirectConnectSoftWhiteBulbA19
    128: DeviceType.LIGHT,
    # DirectConnectTunableWhiteBulbA19
    129: DeviceType.LIGHT,
    # DirectConnectTunableWhiteBulbBR30
    130: DeviceType.LIGHT,
    # DirectConnectFullColorBulbA19
    131: DeviceType.LIGHT,
    # DirectConnectFullColorBulbBR30
    132: DeviceType.LIGHT,
    # DirectConnectFullColorLightStrip
    133: DeviceType.INDOOR_LIGHT_STRIP,
    # SingleChipSoftWhiteA19
    134: DeviceType.LIGHT,
    # SingleChipTunableWhiteA19
    135: DeviceType.LIGHT,
    # SingleChipTunableWhiteBR30
    136: DeviceType.LIGHT,
    # SingleChipFullColorBulbA19
    137: DeviceType.LIGHT,
    # SingleChipFullColorBulbBR30
    138: DeviceType.LIGHT,
    # SingleChipFullColorIndoorValueLightStripGen2
    139: DeviceType.INDOOR_LIGHT_STRIP,
    # SingleChipFullColorBulbPAR38Outdoor
    140: DeviceType.LIGHT,
    # SingleChipFullColorOutdoorPremiumLightStrip16ftGen2
    141: DeviceType.OUTDOOR_LIGHT_STRIP,
    # SingleChipFullColorWafer4InchGen1
    142: DeviceType.DOWNLIGHT,
    # SingleChipFullColorWafer6InchGen1
    143: DeviceType.DOWNLIGHT,
    # SingleChipTunableWhiteWafer4InchGen1
    144: DeviceType.DOWNLIGHT,
    # SingleChipTunableWhiteWafer6InchGen1
    145: DeviceType.DOWNLIGHT,
    # SingleChipFullColorBulbST19Gen1
    146: DeviceType.LIGHT,
    # SingleChipFullColorBulbG25Gen1
    147: DeviceType.LIGHT,
    # SingleChipSoftWhiteBulbST19Gen1
    148: DeviceType.LIGHT,
    # SingleChipSoftWhiteBulbG25Gen1
    149: DeviceType.LIGHT,
    # SingleChipSoftWhiteBulbBCGen1
    150: DeviceType.LIGHT,
    # SingleChipSoftWhiteBulbBMGen1
    151: DeviceType.LIGHT,
    # SingleChipRevealSoftWhiteBulbA19
    152: DeviceType.LIGHT,
    # SingleChipRevealFullColorBulbA19
    153: DeviceType.LIGHT,
    # SingleChipRevealFullColorBulbBR30
    154: DeviceType.LIGHT,
    # SingleChipFullColorDynamicEffectsBulbA19
    155: DeviceType.LIGHT,
    # SingleChipRevealFullColorBulbA21
    156: DeviceType.LIGHT,
    # SingleChipFullColorDynamicEffectsBulbBR30
    157: DeviceType.LIGHT,
    # SingleChipFullColorIndoorPremiumLightStripGen2
    158: DeviceType.INDOOR_LIGHT_STRIP,
    # SingleChipFullColorOutdoorPremiumLightStrip32ftGen2
    159: DeviceType.OUTDOOR_LIGHT_STRIP,
    # SingleChipFullColorBulbA21
    160: DeviceType.LIGHT,
    # SingleChipRevealFullColorWafer2InchGen1
    161: DeviceType.DOWNLIGHT,
    # SingleChipRevealFullColorGimbalWafer4InchGen1
    162: DeviceType.DOWNLIGHT,
    # SingleChipRevealFullColorGimbalWafer6InchGen1
    163: DeviceType.DOWNLIGHT,
    # SingleChipRevealFullColorHighLumenWafer4InchGen1
    164: DeviceType.DOWNLIGHT,
    # SingleChipRevealFullColorHighLumenWafer6InchGen1
    165: DeviceType.DOWNLIGHT,
    # SingleChipFullColorDynamicEffectsNeonLightStrip10ftGen1
    166: DeviceType.NEON_LIGHT_STRIP,
    # SingleChipFullColorDynamicEffectsNeonLightStrip16ftGen1
    167: DeviceType.NEON_LIGHT_STRIP,
    # SingleChipFullColorDynamicEffectsHexagonLightTileGen1
    168: DeviceType.LIGHT_TILE,
    # SingleChipRevealFullColorWafer4InchGen1
    169: DeviceType.DOWNLIGHT,
    # SingleChipRevealFullColorWafer6InchGen1
    170: DeviceType.DOWNLIGHT,
    # SingleChipFullColorBulbA19Gen3
    171: DeviceType.LIGHT,
    # SingleChipPlugTCOGen1
    172: DeviceType.PLUG,
    # SingleChipFullColorBulbBR30Gen3
    173: DeviceType.LIGHT,
    # SingleChipRevealFullColorWafer4InchGen2
    174: DeviceType.DOWNLIGHT,
    # SingleChipRevealFullColorWafer6InchGen2
    175: DeviceType.DOWNLIGHT,
    # SingleChipRevealFullColorWafer2InchGen2
    177: DeviceType.DOWNLIGHT,
    # SingleChipRevealFullColorHighLumenWafer4InchGen2
    180: DeviceType.DOWNLIGHT,
    # SingleChipRevealFullColorHighLumenWafer6InchGen2
    181: DeviceType.DOWNLIGHT,
    # SingleChipFullColorBulbA19ClearSpiralGen1
    182: DeviceType.LIGHT,
    # ThermostatGen1
    224: DeviceType.THERMOSTAT,
    # CameraIndoorGen1
    240: DeviceType.CAMERA,
    # CameraOutdoorWiredGen1
    241: DeviceType.CAMERA,
    # CameraOutdoorBatteryGen1
    242: DeviceType.CAMERA
}


class CyncError(Exception):
    """Cync error."""

class TwoFactorRequiredError(CyncError):
    """Two-factor required."""

class AuthFailedError(CyncError):
    """Auth failed."""

class NoHubConnectedError(CyncError):
    """No hub device is connected to Wi-Fi."""

class MissingAuthError(Exception):
    """Missing auth error."""

class BadRequestError(Exception):
    """For HTTP 400 Responses."""

class UnsupportedCapabilityError(Exception):
    """When a property that the device doesn't support is attempted to be accessed."""


from __future__ import annotations
from typing import Callable, TYPE_CHECKING

from pycync.exceptions import CyncError

if TYPE_CHECKING:
    from pycync.devices.groups import CyncHome

_user_homes: dict[int, UserHomes] = {}


def get_user_homes(user_id: int):
    """Fetch a list of all configured homes for the user."""

    current_homes = _user_homes.get(user_id, UserHomes([]))
    return current_homes.homes


def set_user_homes(user_id: int, homes: list[CyncHome]):
    """Set a list of configured homes for the user."""

    current_homes = _user_homes.get(user_id, UserHomes([]))
    current_homes.homes = homes

    _user_homes[user_id] = current_homes


def get_user_device_callback(user_id: int):
    """Get the configured device update callback function for the user."""

    current_homes = _user_homes.get(user_id, UserHomes([]))
    return current_homes.on_data_update


def set_user_device_callback(user_id: int, callback: Callable):
    """Set the configured device update callback function for the user."""

    current_homes = _user_homes.get(user_id, UserHomes([]))
    current_homes.on_data_update = callback

    _user_homes[user_id] = current_homes


def get_associated_home(user_id: int, device_id: int):
    """Get the home that the provided device id belongs to."""

    user_homes = _user_homes.get(user_id, UserHomes([])).homes

    found_home = next((home for home in user_homes if home.contains_device_id(device_id)), None)
    if found_home is None:
        raise CyncError(f"Device ID {device_id} not found on user account {user_id}.")
    return found_home


def get_associated_home_devices(user_id: int, device_id: int):
    """Given a device ID, returns a list of all devices that exist in the same home that the device ID belongs to."""

    home_for_device = get_associated_home(user_id, device_id)
    return home_for_device.get_flattened_device_list()


def get_flattened_devices(user_id: int):
    """Returns a list of all devices that have been configured for the user, across all homes."""

    homes = _user_homes.get(user_id, UserHomes([])).homes
    all_devices = []

    for home in homes:
        all_devices.extend(home.get_flattened_device_list())

    return all_devices


class UserHomes:
    """
    A summary of all homes associated with a user, and an optional callback function
    to call when any of the home's devices are updated.
    """

    def __init__(self, homes: list[CyncHome], on_data_update: Callable = None):
        self.homes = homes
        self.on_data_update = on_data_update



from pycync.tcp import packet_builder
from tests import TEST_USER_ID

TEST_DEVICE_ID = 23456
TEST_DEVICE_MESH_ID = 5

def test_build_login_packet(mocker):
    mocker.patch("pycync.tcp.packet_builder._get_and_increment_packet_counter", return_value=1)

    test_auth_string = "123456789abcdef"
    auth_string_bytes = bytearray(test_auth_string, "ascii").hex()

    login_packet = packet_builder.build_login_request_packet(test_auth_string, TEST_USER_ID)

    assert login_packet == bytearray.fromhex("1300000019030001e240000f" + auth_string_bytes + "00001e")

def test_build_state_query_packet(mocker):
    mocker.patch("pycync.tcp.packet_builder._get_and_increment_packet_counter", return_value=1)
    mocker.patch("pycync.tcp.inner_packet_builder._get_and_increment_sequence_bytes", return_value=int(257).to_bytes(4, "little"))

    state_query_packet = packet_builder.build_state_query_request_packet(TEST_DEVICE_ID)

    assert state_query_packet == bytearray.fromhex("730000001800005ba00001007e01010000f85206000000ffff0000567e")

def test_build_power_state_request_packet(mocker):
    mocker.patch("pycync.tcp.packet_builder._get_and_increment_packet_counter", return_value=1)
    mocker.patch("pycync.tcp.inner_packet_builder._get_and_increment_sequence_bytes", return_value=int(257).to_bytes(4, "little"))

    power_state_request_packet = packet_builder.build_power_state_request_packet(TEST_DEVICE_ID, TEST_DEVICE_MESH_ID, True)

    assert power_state_request_packet == bytearray.fromhex("730000001f00005ba00001007e01010000f8d00d0001010000000500d01102010000c87e")

def test_build_brightness_request_packet(mocker):
    mocker.patch("pycync.tcp.packet_builder._get_and_increment_packet_counter", return_value=1)
    mocker.patch("pycync.tcp.inner_packet_builder._get_and_increment_sequence_bytes", return_value=int(257).to_bytes(4, "little"))

    brightness_request_packet = packet_builder.build_brightness_request_packet(TEST_DEVICE_ID, TEST_DEVICE_MESH_ID, 42)

    assert brightness_request_packet == bytearray.fromhex("730000001d00005ba00001007e01010000f8d20b0001010000000500d211022af37e")

def test_build_color_temp_request_packet(mocker):
    mocker.patch("pycync.tcp.packet_builder._get_and_increment_packet_counter", return_value=1)
    mocker.patch("pycync.tcp.inner_packet_builder._get_and_increment_sequence_bytes", return_value=int(257).to_bytes(4, "little"))

    color_temp_request_packet = packet_builder.build_color_temp_request_packet(TEST_DEVICE_ID, TEST_DEVICE_MESH_ID, 56)

    assert color_temp_request_packet == bytearray.fromhex("730000001e00005ba00001007e01010000f8e20c0001010000000500e211020538277e")

def test_build_rgb_request_packet(mocker):
    mocker.patch("pycync.tcp.packet_builder._get_and_increment_packet_counter", return_value=1)
    mocker.patch("pycync.tcp.inner_packet_builder._get_and_increment_sequence_bytes", return_value=int(257).to_bytes(4, "little"))

    rgb_request_packet = packet_builder.build_rgb_request_packet(TEST_DEVICE_ID, TEST_DEVICE_MESH_ID, (190, 239, 237))

    assert rgb_request_packet == bytearray.fromhex("730000002000005ba00001007e01010000f8e20e0001010000000500e2110204beefed8a7e")

def test_serialized_7e_packet(mocker):
    mocker.patch("pycync.tcp.packet_builder._get_and_increment_packet_counter", return_value=1)
    mocker.patch("pycync.tcp.inner_packet_builder._get_and_increment_sequence_bytes", return_value=int(257).to_bytes(4, "little"))

    power_state_request_packet = packet_builder.build_power_state_request_packet(TEST_DEVICE_ID, 0x7e, True)

    assert power_state_request_packet == bytearray.fromhex("730000002000005ba00001007e01010000f8d00d0001010000007d5e00d01102010000417e")

def test_sequence_generation():
    power_state_request_packet_1 = packet_builder.build_power_state_request_packet(TEST_DEVICE_ID, TEST_DEVICE_MESH_ID, True)
    power_state_request_packet_2 = packet_builder.build_power_state_request_packet(TEST_DEVICE_ID, TEST_DEVICE_MESH_ID, True)
    power_state_request_packet_3 = packet_builder.build_power_state_request_packet(TEST_DEVICE_ID, TEST_DEVICE_MESH_ID, True)
    power_state_request_packet_4 = packet_builder.build_power_state_request_packet(TEST_DEVICE_ID, TEST_DEVICE_MESH_ID, True)
    power_state_request_packet_5 = packet_builder.build_power_state_request_packet(TEST_DEVICE_ID, TEST_DEVICE_MESH_ID, True)

    assert power_state_request_packet_1 == bytearray.fromhex("730000001f00005ba00001007e01010000f8d00d0001010000000500d01102010000c87e")
    assert power_state_request_packet_2 == bytearray.fromhex("730000001f00005ba00002007e02010000f8d00d0002010000000500d01102010000c97e")
    assert power_state_request_packet_3 == bytearray.fromhex("730000001f00005ba00003007e03010000f8d00d0003010000000500d01102010000ca7e")
    assert power_state_request_packet_4 == bytearray.fromhex("730000001f00005ba00004007e04010000f8d00d0004010000000500d01102010000cb7e")
    assert power_state_request_packet_5 == bytearray.fromhex("730000001f00005ba00005007e05010000f8d00d0005010000000500d01102010000cc7e")


"""
The proverbial "beating heart" of the Cync client.
This client listens for device state changes and updates them accordingly, and also handles sending all device action commands.
"""

from __future__ import annotations
from typing import TYPE_CHECKING

import asyncio
import logging

from .packet import MessageType, ParsedMessage, PipeCommandCode
from .tcp_manager import TcpManager
from pycync.devices.controllable import CyncControllable
from pycync.exceptions import NoHubConnectedError, CyncError
from pycync.devices.capabilities import CyncCapability
from pycync.devices import device_storage
from pycync.user import User

if TYPE_CHECKING:
    from pycync.devices import CyncDevice
    from pycync.devices.groups import CyncHome


class CommandClient:
    _LOGGER = logging.getLogger(__name__)

    def __init__(self, user: User):
        self._user = user

        self._device_statuses_updated = False
        self._tcp_manager: TcpManager = None

    def start_connection(self):
        self._tcp_manager = TcpManager(self._user, self.on_message_received)

    def on_message_received(self, parsed_message: ParsedMessage):
        match parsed_message.message_type:
            case MessageType.LOGIN.value:
                self.probe_devices()
            case MessageType.PROBE.value if parsed_message.version != 0:
                devices_in_home = device_storage.get_associated_home_devices(self._user.user_id,
                                                                             parsed_message.device_id)
                device = next(device for device in devices_in_home if device.device_id == parsed_message.device_id)
                device.set_wifi_connected(True)
                self._device_statuses_updated = True
            case MessageType.SYNC.value:
                callback = device_storage.get_user_device_callback(self._user.user_id)
                if callback is not None:
                    callback(parsed_message.data)
            case MessageType.PIPE.value:
                if parsed_message.command_code == PipeCommandCode.QUERY_DEVICE_STATUS_PAGES.value:
                    callback = device_storage.get_user_device_callback(self._user.user_id)
                    if callback is not None:
                        callback(parsed_message.data)

    def probe_devices(self):
        self._tcp_manager.probe_devices(device_storage.get_flattened_devices(self._user.user_id))

    async def update_mesh_devices(self):
        """Get new device state."""
        homes_for_user = device_storage.get_user_homes(self._user.user_id)

        hub_devices: list[CyncDevice] = []
        for home in homes_for_user:
            hub_device = await self._fetch_hub_device(home)
            hub_devices.append(hub_device)

        self._tcp_manager.update_mesh_devices(hub_devices)

    async def set_power_state(self, controllable: CyncControllable, is_on: bool):
        """Set device(s) to either on or off."""
        hub_device = await self._fetch_hub_device(controllable.parent_home)

        self._tcp_manager.set_power_state(hub_device, controllable.mesh_reference_id, is_on)

    async def set_brightness(self, controllable: CyncControllable, brightness: int):
        """Sets the brightness. Must be between 0 and 100 inclusive."""
        if brightness < 0 or brightness > 100:
            raise CyncError("Brightness must be between 0 and 100 inclusive")

        hub_device = await self._fetch_hub_device(controllable.parent_home)

        self._tcp_manager.set_brightness(hub_device, controllable.mesh_reference_id, brightness)

    async def set_color_temp(self, controllable: CyncControllable, color_temp: int):
        """
        Sets the color temperature. Must be between 1 and 100 inclusive.
        1 represents the most "blue" and 100 represents the most "orange".
        """
        if color_temp < 1 or color_temp > 100:
            raise CyncError("Color temperature must be between 1 and 100 inclusive.")

        hub_device = await self._fetch_hub_device(controllable.parent_home)

        self._tcp_manager.set_color_temp(hub_device, controllable.mesh_reference_id, color_temp)

    async def set_rgb(self, controllable: CyncControllable, rgb: tuple[int, int, int]):
        """Sets the RGB color. Each color must be between 0 and 255 inclusive."""
        if rgb[0] > 255 or rgb[1] > 255 or rgb[2] > 255:
            raise CyncError("Each RGB value must be between 0 and 255 inclusive")

        hub_device = await self._fetch_hub_device(controllable.parent_home)

        self._tcp_manager.set_rgb(hub_device, controllable.mesh_reference_id, rgb)

    def shut_down(self):
        self._tcp_manager.shut_down()

    async def _fetch_hub_device(self, home: CyncHome) -> CyncDevice:
        """
        Fetches an eligible 'hub device' from a given home.
        A hub device is a device that is actively connected to Wi-Fi, and can act as a proxy into the Bluetooth mesh.
        """

        while not self._device_statuses_updated:
            await asyncio.sleep(1)
            self._LOGGER.debug("Awaiting probe initialization before fetching hub.")

        hub_device = next((device for device in home.get_flattened_device_list() if
                           device.wifi_connected and CyncCapability.SIG_MESH in device.capabilities), None)
        if hub_device is None:
            raise NoHubConnectedError

        return hub_device



"""
Module responsible for building 'inner' packet frames, typically seen on PIPE and PIPE_SYNC type TCP messages.
Because not every packet includes an inner frame, and the fact that these inner packets use a different
endianness, constructing these inner frames is handled in this separate module.
"""

import threading

from .packet import PipeCommandCode, generate_checksum, generate_zero_bytes

_INNER_PACKET_DELIMITER = bytearray.fromhex("7e")

_inner_packet_counter = 257  # Starts at 0x0101
_inner_packet_counter_lock = threading.Lock()


def build_query_device_inner_packet(pipe_direction):
    sequence_bytes = _get_and_increment_sequence_bytes()
    packet_direction_bytes = pipe_direction.to_bytes(1, "little")

    limit = bytearray.fromhex("ffff")  # Update all devices
    offset = bytearray.fromhex("0000")  # Start at the beginning

    command_bytes = generate_zero_bytes(2) + limit + offset

    return _compile_final_packet(sequence_bytes, packet_direction_bytes,
                                 PipeCommandCode.QUERY_DEVICE_STATUS_PAGES.value, command_bytes)


def build_power_state_inner_packet(pipe_direction, standalone_mesh_id, is_on):
    command_code = PipeCommandCode.SET_POWER_STATE.value
    sequence_bytes = _get_and_increment_sequence_bytes()
    packet_direction_bytes = pipe_direction.to_bytes(1, "little")

    mesh_id_bytes = standalone_mesh_id.to_bytes(2, 'little')
    command_code_bytes = command_code.to_bytes(1, "little")
    extra_command_bytes = bytearray.fromhex("1102")
    is_on_bytes = is_on.to_bytes(1, "little")

    command_bytes = (generate_zero_bytes(1) +
                     mesh_id_bytes +
                     command_code_bytes +
                     extra_command_bytes +
                     is_on_bytes +
                     generate_zero_bytes(2))

    return _compile_final_packet(sequence_bytes, packet_direction_bytes, command_code, command_bytes)


def build_brightness_inner_packet(pipe_direction, standalone_mesh_id, brightness):
    command_code = PipeCommandCode.SET_BRIGHTNESS.value
    sequence_bytes = _get_and_increment_sequence_bytes()
    packet_direction_bytes = pipe_direction.to_bytes(1, "little")

    mesh_id_bytes = standalone_mesh_id.to_bytes(2, 'little')
    command_code_bytes = command_code.to_bytes(1, "little")
    extra_command_bytes = bytearray.fromhex("1102")
    brightness_bytes = brightness.to_bytes(1, "little")

    command_bytes = (generate_zero_bytes(1) +
                     mesh_id_bytes +
                     command_code_bytes +
                     extra_command_bytes +
                     brightness_bytes)

    return _compile_final_packet(sequence_bytes, packet_direction_bytes, command_code, command_bytes)


def build_color_temp_inner_packet(pipe_direction, standalone_mesh_id, color_temp):
    command_code = PipeCommandCode.SET_COLOR.value
    sequence_bytes = _get_and_increment_sequence_bytes()
    packet_direction_bytes = pipe_direction.to_bytes(1, "little")

    mesh_id_bytes = standalone_mesh_id.to_bytes(2, 'little')
    command_code_bytes = command_code.to_bytes(1, "little")
    extra_command_bytes = bytearray.fromhex("110205")
    color_temp_bytes = color_temp.to_bytes(1, "little")

    command_bytes = (generate_zero_bytes(1) +
                     mesh_id_bytes +
                     command_code_bytes +
                     extra_command_bytes +
                     color_temp_bytes)

    return _compile_final_packet(sequence_bytes, packet_direction_bytes, command_code, command_bytes)


def build_rgb_inner_packet(pipe_direction, standalone_mesh_id, rgb: tuple[int, int, int]):
    command_code = PipeCommandCode.SET_COLOR.value
    sequence_bytes = _get_and_increment_sequence_bytes()
    packet_direction_bytes = pipe_direction.to_bytes(1, "little")

    mesh_id_bytes = standalone_mesh_id.to_bytes(2, 'little')
    command_code_bytes = command_code.to_bytes(1, "little")
    extra_command_bytes = bytearray.fromhex("110204")
    rgb_bytes = bytearray(rgb)

    command_bytes = (generate_zero_bytes(1) +
                     mesh_id_bytes +
                     command_code_bytes +
                     extra_command_bytes +
                     rgb_bytes)

    return _compile_final_packet(sequence_bytes, packet_direction_bytes, command_code, command_bytes)


def _compile_final_packet(sequence_bytes, pipe_direction_bytes, pipe_command_code, command_bytes) -> bytearray:
    command_code_bytes = pipe_command_code.to_bytes(1, "little")

    if _requires_second_sequence_inserted(pipe_command_code):
        command_bytes = sequence_bytes + command_bytes

    packet_command_arguments_length = len(command_bytes).to_bytes(2, "little")

    packet_command_body = command_code_bytes + packet_command_arguments_length + command_bytes
    checksum = generate_checksum(packet_command_body).to_bytes(1, "little")

    packet_body = (sequence_bytes +
                   pipe_direction_bytes +
                   packet_command_body +
                   checksum)

    encoded_body = _encode_7e_usages(packet_body)

    return _INNER_PACKET_DELIMITER + encoded_body + _INNER_PACKET_DELIMITER


def _requires_second_sequence_inserted(pipe_command):
    return pipe_command in [
        PipeCommandCode.SET_POWER_STATE.value,
        PipeCommandCode.SET_COLOR.value,
        PipeCommandCode.SET_BRIGHTNESS.value,
        PipeCommandCode.COMBO_CONTROL.value
    ]


def _get_and_increment_sequence_bytes():
    with _inner_packet_counter_lock:
        global _inner_packet_counter
        counter_value = _inner_packet_counter
        _inner_packet_counter = _inner_packet_counter + 1 if _inner_packet_counter + 1 < 4294967295 else 257

        return counter_value.to_bytes(4, "little")


def _encode_7e_usages(frame_bytes: bytearray) -> bytearray:
    """
    When sending inner frames, the byte 0x7e is encoded as 0x7d5e if it's within the inner frame,
    so it isn't mistaken for a frame boundary marker.
    """
    return frame_bytes.replace(b"\x7e", b"\x7d\x5e")



"""Module responsible for parsing raw TCP packets received from the Cync server into a more usable format."""

from __future__ import annotations
from typing import TYPE_CHECKING

import struct

from ..devices import device_storage
from .packet import ParsedMessage, ParsedInnerFrame, MessageType, PipeCommandCode, generate_checksum
from pycync.devices.capabilities import DEVICE_CAPABILITIES, CyncCapability

if TYPE_CHECKING:
    from pycync.devices import CyncDevice


def parse_packet(packet: bytearray, user_id: int) -> ParsedMessage:
    packet_type = (packet[0] & 0xF0) >> 4
    is_response = bool((packet[0] & 0x08) >> 3)
    version = packet[0] & 0x7
    packet_length = struct.unpack(">I", packet[1:5])[0]
    packet = packet[5:]

    if len(packet) != packet_length:
        raise ValueError(
            "Provided packet length did not match actual packet length. Expected: {}, got: {}".format(packet_length,
                                                                                                      len(packet)))

    match packet_type:
        case MessageType.LOGIN.value:
            return ParsedMessage(MessageType.LOGIN.value, is_response, None, None, version)
        case MessageType.PROBE.value:
            return _parse_probe_packet(packet, is_response, version)
        case MessageType.SYNC.value:
            return _parse_sync_packet(packet, is_response, version, user_id)
        case MessageType.PIPE.value:
            return _parse_pipe_packet(packet, packet_length, is_response, version, user_id)
        case MessageType.DISCONNECT.value:
            return ParsedMessage(MessageType.DISCONNECT.value, is_response, user_id, packet[0], version)
        case _:
            raise NotImplementedError


def _parse_probe_packet(packet: bytearray, is_response, version) -> ParsedMessage:
    device_id = struct.unpack(">I", packet[0:4])[0]
    data = packet[4:]

    return ParsedMessage(MessageType.PROBE.value, is_response, device_id, data, version)


def _parse_sync_packet(packet: bytearray, is_response, version, user_id) -> ParsedMessage:
    device_id = struct.unpack(">I", packet[0:4])[0]
    device_list = device_storage.get_associated_home_devices(user_id, device_id)
    device_type = next(device.device_type for device in device_list if device.device_id == device_id)
    is_mesh_device = CyncCapability.SIG_MESH in DEVICE_CAPABILITIES[device_type]

    updated_device_data = {}

    if packet[4:7].hex() == '010106' and is_mesh_device:
        packet = packet[7:]
        while len(packet) > 3:
            info_length = struct.unpack(">H", packet[1:3])[0]
            packet = packet[3:info_length + 3]
            mesh_id = packet[0]
            try:
                resolved_device = next(device for device in device_list if device.isolated_mesh_id == mesh_id)
                resolved_device.update_state(bool(packet[1]), packet[2], packet[3], (packet[4], packet[5], packet[6]))
            except StopIteration as ex:
                raise ValueError("Unable to resolve device ID for mesh ID: {}".format(mesh_id)) from ex

            updated_device_data[resolved_device.device_id] = resolved_device
            packet = packet[info_length + 1:]

        return ParsedMessage(MessageType.SYNC.value, is_response, device_id, updated_device_data, version)

    else:
        raise NotImplementedError


def _parse_pipe_packet(packet: bytearray, length, is_response, version, user_id) -> ParsedMessage:
    device_id = struct.unpack(">I", packet[0:4])[0]
    device_list = device_storage.get_associated_home_devices(user_id, device_id)

    if length > 7 and packet[7] == 0x7e:
        inner_frame = _parse_inner_packet_frame(packet[7:], device_list)
    else:
        raise NotImplementedError

    return ParsedMessage(MessageType.PIPE.value, is_response, device_id, inner_frame.data, version,
                         inner_frame.command_type)


def _parse_inner_packet_frame(frame_bytes: bytearray, device_list) -> ParsedInnerFrame:
    if frame_bytes[0] != 0x7e or frame_bytes[-1] != 0x7e:
        raise ValueError("Invalid delimiters for inner packet frame")

    frame_bytes = frame_bytes[1:-1]  # Trim off delimiters
    frame_bytes = _decode_7e_usages(frame_bytes)

    frame_bytes = frame_bytes[4:]  # Trim off sequence number, we don't need it

    command_code = frame_bytes[1]
    data_length = struct.unpack("<H", frame_bytes[2:4])[0]

    frame_checksum = frame_bytes[-1]
    if not _does_checksum_match(frame_bytes[1:-1], frame_checksum):
        raise ValueError("Invalid checksum for inner packet frame")

    match command_code:
        case PipeCommandCode.QUERY_DEVICE_STATUS_PAGES.value:
            parsed_data = _parse_device_status_pages_command(frame_bytes[4: 4 + data_length], device_list)
        case _:
            raise NotImplementedError

    return ParsedInnerFrame(command_code, parsed_data)


def _parse_device_status_pages_command(data_bytes: bytearray, device_list) -> dict[int, CyncDevice]:
    updated_device_data = {}
    if len(data_bytes) < 5:
        return updated_device_data

    device_count = struct.unpack("<H", data_bytes[4:6])[0]
    trimmed_bytes = data_bytes[6:]

    for i in range(device_count):
        device_data = trimmed_bytes[0:24]

        mesh_id = struct.unpack("<H", device_data[0:2])[0]
        is_online = device_data[3]
        is_on = device_data[8]
        brightness = device_data[12]
        color_mode = device_data[16]
        rgb = (device_data[20], device_data[21], device_data[22])

        try:
            resolved_device = next(device for device in device_list if device.isolated_mesh_id == mesh_id)
            resolved_device.update_state(bool(is_on), brightness, color_mode, rgb, bool(is_online))
        except StopIteration as ex:
            raise ValueError("Unable to resolve device ID for mesh ID: {}".format(mesh_id)) from ex

        updated_device_data[resolved_device.device_id] = resolved_device
        trimmed_bytes = trimmed_bytes[24:]

    return updated_device_data


def _decode_7e_usages(frame_bytes: bytearray) -> bytearray:
    """
    When sending inner frames, the byte 0x7e is encoded as 0x7d5e if it's within the inner frame,
    so it isn't mistaken for a frame boundary marker.
    We need to undo that when reading it.
    """
    return frame_bytes.replace(b"\x7d\x5e", b"\x7e")


def _does_checksum_match(data_bytes: bytearray, expected_checksum: int) -> bool:
    checksum_result = generate_checksum(data_bytes)
    return checksum_result == expected_checksum



from __future__ import annotations
from typing import TYPE_CHECKING

from abc import abstractmethod
from typing import Protocol

from pycync.devices.capabilities import CyncCapability

if TYPE_CHECKING:
    from pycync.devices.groups import CyncHome


class CyncControllable(Protocol):
    """Protocol describing any Cync entity that can be controlled by the user."""

    parent_home: CyncHome

    @property
    @abstractmethod
    def capabilities(self) -> frozenset[CyncCapability]:
        pass

    @property
    @abstractmethod
    def name(self) -> str:
        pass

    @property
    @abstractmethod
    def mesh_reference_id(self) -> int:
        pass

    @property
    @abstractmethod
    def unique_id(self) -> str:
        """Provides an identifier that uniquely identifies this controllable entity."""
        pass

    @abstractmethod
    def supports_capability(self, capability: CyncCapability) -> bool:
        pass



"""
Contains auth information about an authenticated user, and performs generic REST API calls for the user.
"""
import time
from asyncio import TimeoutError
from json import dumps
from typing import Any

from aiohttp import ClientSession, ClientResponseError

from .const import GE_CORP_ID, REST_API_BASE_URL
from .user import User
from .exceptions import BadRequestError, TwoFactorRequiredError, AuthFailedError, CyncError


class Auth:
    def __init__(self, session: ClientSession, user: User = None, username: str = None, password: str = None) -> None:
        """Initialize the auth."""
        self._session = session
        self._user = user
        self._username = username
        self._password = password

    @property
    def user(self):
        return self._user

    @property
    def session(self):
        return self._session

    @property
    def username(self):
        return self._username

    @property
    def password(self):
        return self._password

    async def login(self, two_factor_code: str = None) -> User:
        """
        Attempts to log in with the configured user information.

        If a two factor code is not provided, and the server requests a two factor code, a TwoFactorRequiredError
        will be raised, and at the same time a two factor code will be sent to the user's email.
        This function can then be called again with the two factor code provided to authenticate with the code.
        """
        if two_factor_code is None:
            try:
                user_info = await self._async_auth_user()
                self._user = User(user_info["access_token"], user_info["refresh_token"], user_info["authorize"],
                                  user_info["user_id"], expire_in=user_info["expire_in"])
                return self._user
            except TwoFactorRequiredError:
                two_factor_request = {'corp_id': GE_CORP_ID, 'email': self.username, 'local_lang': "en-us"}

                await self._send_user_request(url=f'{REST_API_BASE_URL}/v2/two_factor/email/verifycode', method="POST",
                                              json=two_factor_request)
                raise TwoFactorRequiredError('Two factor verification required. Code sent to user email.')
        else:
            user_info = await self._async_auth_user_two_factor(two_factor_code)
            self._user = User(user_info["access_token"], user_info["refresh_token"], user_info["authorize"],
                              user_info["user_id"], expire_in=user_info["expire_in"])
            return self._user

    async def _async_auth_user(self):
        """Attempt to authenticate user without a two factor code."""
        auth_data = {'corp_id': GE_CORP_ID, 'email': self.username, 'password': self.password}

        try:
            auth_response = await self._send_user_request(url=f'{REST_API_BASE_URL}/v2/user_auth', method="POST",
                                                          json=auth_data)
            return auth_response
        except BadRequestError:
            raise TwoFactorRequiredError("Two factor verification required.")

    async def _async_auth_user_two_factor(self, two_factor_code: str):
        """Attempt to authenticate user with a two factor code."""
        two_factor_request = {'corp_id': GE_CORP_ID, 'email': self.username, 'password': self.password,
                              'two_factor': two_factor_code, 'resource': 1}

        try:
            auth_response = await self._send_user_request(url=f'{REST_API_BASE_URL}/v2/user_auth/two_factor',
                                                          method="POST", json=two_factor_request)
            return auth_response
        except BadRequestError as ex:
            raise AuthFailedError("Invalid two-factor code") from ex

    async def async_refresh_user_token(self):
        """
        Refresh the user's session token. If the token has already expired, a new login will be required.
        (Likely also requiring a new two factor code to be provided.)
        """
        refresh_request = {'refresh_token': self._user.refresh_token}

        try:
            body = dumps(refresh_request)

            resp = await self.session.request(method="POST", url=f'{REST_API_BASE_URL}/v2/user/token/refresh',
                                              data=body)
            if resp.status != 200:
                raise AuthFailedError('Refresh token failed')

            auth_response = await resp.json()

            self._user.set_new_access_token(auth_response["access_token"], auth_response["refresh_token"],
                                            auth_response["expire_in"])
        except Exception as ex:
            raise AuthFailedError(ex)

    async def _send_user_request(
            self,
            url: str,
            method: str = "GET",
            json: dict[Any, Any] | None = None,
            raise_for_status: bool = True,
    ) -> dict:
        """Send an HTTP request with the provided parameters."""
        headers = {}
        if self.user:
            if self.user.expires_at - time.time() < 3600:
                await self.async_refresh_user_token()
            headers["Access-Token"] = self.user.access_token

        try:
            if json:
                body = dumps(json)

                resp = await self.session.request(method, url, headers=headers, data=body)
            else:
                resp = await self.session.request(method, url, headers=headers)

        except TimeoutError as ex:
            msg = f"Timeout error during query of url {url}: {ex}"
            raise CyncError(msg) from ex
        except Exception as ex:
            msg = f"Unknown error during query of url {url}: {ex}"
            raise CyncError(msg) from ex

        async with resp:
            if resp.status == 400:
                raise BadRequestError("Bad Request")

            if resp.status == 401 or resp.status == 403:
                await self.async_refresh_user_token()

                headers["Access-Token"] = self.user.access_token

                if json:
                    body = dumps(json)

                    resp = await self.session.request(method, url, headers=headers, data=body)
                else:
                    resp = await self.session.request(method, url, headers=headers)

            if raise_for_status:
                try:
                    resp.raise_for_status()
                except ClientResponseError as ex:
                    msg = (
                        f"HTTP error with status code {resp.status} "
                        f"during query of url {url}: {ex}"
                    )
                    raise CyncError(msg) from ex

            return await resp.json()



import time


class User:
    """A Cync user account and their auth credentials."""

    def __init__(self, access_token: str, refresh_token: str, authorize: str, user_id: int, expire_in: int = None,
                 expires_at: float = None):
        """Initialize the User object."""
        self._access_token = access_token
        self._refresh_token = refresh_token
        self._authorize = authorize
        self._user_id = user_id
        self._expires_at = time.time() + expire_in if not expires_at else expires_at

    @property
    def access_token(self) -> str:
        """Return the user's access token."""
        return self._access_token

    @property
    def refresh_token(self) -> str:
        """Return the user's refresh token."""
        return self._refresh_token

    @property
    def expires_at(self) -> float:
        """Return the epoch millis at which the token expires."""
        return self._expires_at

    @property
    def authorize(self) -> str:
        """Return the "authorize" string used when authenticating the TCP connection."""
        return self._authorize

    @property
    def user_id(self) -> int:
        """Return the ID of the user."""
        return self._user_id

    def set_new_access_token(self, access_token: str, refresh_token: str, expire_in: float):
        """Set the new access token."""
        self._access_token = access_token
        self._refresh_token = refresh_token
        self._expires_at = time.time() + expire_in



import pytest

from pycync import CyncHome, CyncDevice, CyncGroup, CyncRoom
from pycync.devices import device_storage
from pycync.exceptions import CyncError
from tests import TEST_USER_ID


home_1 = CyncHome("Home 1", 1234, [], [])
home_2 = CyncHome("Home 2", 2345, [], [])

home_1_room_1_group_1_device_1 = CyncDevice({"id": 12}, {"deviceID": 2, "deviceType": 224}, home_1, None, True)
home_1_room_1_group_1_device_2 = CyncDevice({"id": 23}, {"deviceID": 2, "deviceType": 224}, home_1, None, True)
home_1_room_1_group_1 = CyncGroup("Group 1", 1, home_1, [home_1_room_1_group_1_device_1, home_1_room_1_group_1_device_2], None)

home_1_room_1_device_1 = CyncDevice({"id": 34}, {"deviceID": 2, "deviceType": 224}, home_1, None, True)
home_1_room_1 = CyncRoom("Room 1", 2, home_1, [home_1_room_1_group_1], [home_1_room_1_device_1], None)

home_1_room_2_device_1 = CyncDevice({"id": 45}, {"deviceID": 2, "deviceType": 224}, home_1, None, True)
home_1_room_2 = CyncRoom("Room 2", 3, home_1, [], [home_1_room_2_device_1], None)

home_2_room_1_device_1 = CyncDevice({"id": 56}, {"deviceID": 2, "deviceType": 224}, home_2, None, True)
home_2_room_1_device_2 = CyncDevice({"id": 67}, {"deviceID": 2, "deviceType": 224}, home_2, None, True)
home_2_room_1 = CyncRoom("Room 1", 4, home_2, [], [home_2_room_1_device_1, home_2_room_1_device_2], None)

home_2_device_1 = CyncDevice({"id": 78}, {"deviceID": 2, "deviceType": 224}, home_2, None, True)
home_2_device_2 = CyncDevice({"id": 89}, {"deviceID": 2, "deviceType": 224}, home_2, None, True)

home_1.rooms = [home_1_room_1, home_1_room_2]
home_2.rooms = [home_2_room_1]
home_2.global_devices = [home_2_device_1, home_2_device_2]

def test_get_user_homes_no_entries():
    homes: list[CyncHome] = device_storage.get_user_homes(TEST_USER_ID)

    assert len(homes) == 0

def test_set_and_get_user_devices():
    device_storage.set_user_homes(TEST_USER_ID, [home_1, home_2])

    all_devices = device_storage.get_flattened_devices(TEST_USER_ID)

    expected_devices = [
        home_1_room_1_group_1_device_1,
        home_1_room_1_group_1_device_2,
        home_1_room_1_device_1,
        home_1_room_2_device_1,
        home_2_room_1_device_1,
        home_2_room_1_device_2,
        home_2_device_1,
        home_2_device_2
    ]

    result_difference = set(all_devices).difference(set(expected_devices))
    assert not result_difference

def test_set_and_get_user_homes():
    device_storage.set_user_homes(TEST_USER_ID, [home_1, home_2])

    all_homes = device_storage.get_user_homes(TEST_USER_ID)

    expected_homes = [
        home_1,
        home_2
    ]

    result_difference = set(all_homes).difference(set(expected_homes))
    assert not result_difference

def test_get_callback_not_set():
    callback = device_storage.get_user_device_callback(TEST_USER_ID)

    assert callback is None

def test_set_and_get_callback():
    def my_callback():
        return 42

    device_storage.set_user_device_callback(TEST_USER_ID, my_callback)

    callback = device_storage.get_user_device_callback(TEST_USER_ID)

    assert callback == my_callback
    assert callback() == 42

def test_get_associated_home():
    device_storage.set_user_homes(TEST_USER_ID, [home_1, home_2])

    home_1_test_1 = device_storage.get_associated_home(TEST_USER_ID, home_1_room_1_group_1_device_1.device_id)
    home_1_test_2 = device_storage.get_associated_home(TEST_USER_ID, home_1_room_1_device_1.device_id)
    home_1_test_3 = device_storage.get_associated_home(TEST_USER_ID, home_1_room_2_device_1.device_id)
    home_2_test_1 = device_storage.get_associated_home(TEST_USER_ID, home_2_room_1_device_1.device_id)
    home_2_test_2 = device_storage.get_associated_home(TEST_USER_ID, home_2_device_1.device_id)

    assert home_1_test_1 == home_1
    assert home_1_test_2 == home_1
    assert home_1_test_3 == home_1
    assert home_2_test_1 == home_2
    assert home_2_test_2 == home_2

def test_get_associated_home_device_not_found():
    device_storage.set_user_homes(TEST_USER_ID, [home_1, home_2])

    with pytest.raises(CyncError,
                       match=f'Device ID 9876 not found on user account {TEST_USER_ID}.'):
        device_storage.get_associated_home(TEST_USER_ID, 9876)

def test_get_devices_in_associated_home():
    device_storage.set_user_homes(TEST_USER_ID, [home_1, home_2])

    home_1_devices = [
        home_1_room_1_group_1_device_1,
        home_1_room_1_group_1_device_2,
        home_1_room_1_device_1,
        home_1_room_2_device_1
    ]

    home_2_devices = [
        home_2_room_1_device_1,
        home_2_room_1_device_2,
        home_2_device_1,
        home_2_device_2
    ]

    home_1_test_1 = device_storage.get_associated_home_devices(TEST_USER_ID, home_1_room_1_group_1_device_1.device_id)
    home_1_test_2 = device_storage.get_associated_home_devices(TEST_USER_ID, home_1_room_1_device_1.device_id)
    home_1_test_3 = device_storage.get_associated_home_devices(TEST_USER_ID, home_1_room_2_device_1.device_id)
    home_2_test_1 = device_storage.get_associated_home_devices(TEST_USER_ID, home_2_room_1_device_1.device_id)
    home_2_test_2 = device_storage.get_associated_home_devices(TEST_USER_ID, home_2_device_1.device_id)

    assert not set(home_1_test_1).difference(set(home_1_devices))
    assert not set(home_1_test_2).difference(set(home_1_devices))
    assert not set(home_1_test_3).difference(set(home_1_devices))
    assert not set(home_2_test_1).difference(set(home_2_devices))
    assert not set(home_2_test_2).difference(set(home_2_devices))